/*****************************************************************************/
//
// FILE        :  Image_wav.cc
// AUTHOR      :  Jacco Vink 
// DESCRIPTION :  Implementation of Image class: those functions
//                needed for wavelet transforms
//
//
// COPYRIGHT   :  Jacco Vink, AIP, 1999
//
/*****************************************************************************/
#include "Image.h"


/*****************************************************************************/
//
// FUNCTION    : maxima()
// DESCRIPTION : This function searches for a local maximum starting from 
//               pixel (*imax, *jmax ). The coordinates of the maximum is 
//               stored in (*imax, *jmax) plus the value of the pixel.
//               A maxium is that pixel sticking out above its 8 neigbouring
//               pixels. If no local maxima found, *imax & *jmax are set to 0
//               Follows fits pixel numbering convention.
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//  *imax, *jmax        I/O                 I: starting point
//                                          O: local maxium
//  *val                O                   value of local maximum
//
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::maxima( unsigned int *imax, unsigned int *jmax, float *val )
{
  // Local maxima are not defined near the edges, skip the edges
  *imax = MAX(2,*imax);
  *jmax = MAX(2,*jmax);

  if( nx < 3 || ny < 3 || *imax > nx || *jmax > ny){
    //    warning("maxima(): Pixel minimum is not defined!");
    return FAILURE;
  }// end if
  
  unsigned int i, j;
  int k, l, m, istart;
  bool foundmax = false;


  *imax -= 1; // fits convention
  *jmax -= 1;


  for( j = *jmax ; j < ny-1 && foundmax == false; j++){

    if( j == *jmax )
      istart = *imax;
    else
      istart = 0;

    for( i = istart ; i < nx-1 && foundmax == false; i++){
      *val = pixel[ j*nx + i];

      m = 0;
      for( l = -1; l <= 1 ; l++ )
	for( k = -1; k <= 1 ; k++ ){
	  if( !(l == 0 && k == 0) &&  pixel[ (j+l)*nx + i+k] > *val)
	    m++;

	}//for k

      if( m ==  0 ){
	foundmax = true;
	*imax = i+1; // fits convention
	*jmax = j+1;
      }//if

    }// for i
  }// for( j 


  if( foundmax == false ){
    *val  = 0.0;
    return FAILURE;
  }


  return SUCCESS;
  
}// maxima()


/*****************************************************************************/
//
// FUNCTION    : mhfilter()
// DESCRIPTION : 
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
// sigma               I                         width of the gaussian
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::mhfilter(const Image &im, float  sigmax, float sigmay,
			    int i1, int j1,int i2,int j2 )
{
  if( sigmax > 16.0 || sigmay > 16.0 || sigmax <= 0.2 || sigmay <= 0.2
      || im.nx < 2 || im.ny < 2 ){
    //    warning("mhfilter(): The image is too small or sigma too big!");
    return FAILURE;
  }
  // smaller values of n  make the code run fast but less accurate 
  //  int n = (int)(6.0 * MAX(sigmax,sigmay) + 0.5);
  int n = (int)(4.0 * MAX(sigmax,sigmay) + 1.51);
  //  int n = (int)(4.0 * MAX(sigmax,sigmay) + 1.01);

  float *window;

  window = new float [n*n];
  if( window == 0 ){
    return FAILURE;
  }// window

  // Make a 2D Mexican Hat
  float fac2 = sqrt(M_PI/2.0), fac1 = M_PI *sigmax*sigmay;
  float s2sigmax =sqrt(2.0)*sigmax, s2sigmay =sqrt(2.0)*sigmay;

  float xerf, yerf, xdiff, ydiff;
  float norm= 1.0/(2.0 * M_PI * sigmax * sigmay);
  float errf( double x );
  double negsum=0.0, possum=0.0, sum=0.0,  x1, x2, y1, y2;
  //  float erff(float x);
    float gammp(float a, float x);

  int i;
  for(int j = 0 ; j < n ; j++){
    for( i = 0; i < n ; i++){
      x1 = (float)i - 0.5;
      x2 = (float)i + 0.5;
      y1 = (float)j - 0.5;
      y2 = (float)j + 0.5;
      /*
      xerf = errf( x2/s2sigmax ) -  errf( x1/s2sigmax );
      yerf = errf( y2/s2sigmay ) -  errf( y1/s2sigmay );
      */
      
      if(i != 0) xerf= gammp(0.5,SQR(x2/s2sigmax))-gammp(0.5,SQR(x1/s2sigmax));
      else 	 xerf= gammp(0.5,SQR(x2/s2sigmax))+gammp(0.5,SQR(x1/s2sigmax));
      if(j != 0) yerf= gammp(0.5,SQR(y2/s2sigmay))-gammp(0.5,SQR(y1/s2sigmay));
      else       yerf= gammp(0.5,SQR(y2/s2sigmay))+gammp(0.5,SQR(y1/s2sigmay));
      
      xdiff = x1*exp( -x1*x1/(s2sigmax*s2sigmax)) - 
	      x2*exp( -x2*x2/(s2sigmax*s2sigmax));
      ydiff = y1*exp( -y1*y1/(s2sigmay*s2sigmay)) - 
	      y2*exp( -y2*y2/(s2sigmay*s2sigmay));

      window[j*n + i]= norm * (fac1 * xerf*yerf - 
			       fac2*(sigmay*yerf*(xdiff + fac2*sigmax*xerf) + 
				     sigmax*xerf*(ydiff + fac2*sigmay*yerf)) );

      sum += window[ j*n + i];
      if( window[j*n + i] < 0.0 )
	negsum +=  window[j*n + i];
      else
	possum +=  window[j*n + i];

      // window describes only a quarter of the conv. function
      // correct for the other three quadrants:

      if( i != 0 && j!=0 ){
	sum += 3.0 * window[ j*n + i];
	if( window[j*n + i] < 0.0 )
	  negsum +=  3.0*window[j*n + i];
	else
	  possum +=  3.0*window[j*n + i];
      }
      if( (i==0 && j!=0) || (i!=0 && j==0)){
	sum += window[ j*n + i];
	if( window[j*n + i] < 0.0 )
	  negsum +=  window[j*n + i];
	else
	  possum +=  window[j*n + i];
      }

    }// 
  }//for

  /*
    printf(" sum = %e\n",sum);
    printf("+sum = %e\n",possum);
    printf("-sum = %e\n",negsum);
  */

  // Apply a correction factor to make the normalization zero
  if( possum > 0.0 ){
    float corrfac = float(possum/fabs(negsum));

    //    printf("corrfac = %e\n",corrfac);
    sum = 0.0;
    for(int j = 0 ; j < n ; j++){
      for( i = 0; i < n ; i++){
	if( window[j*n + i] < 0.0 )
	  window[j*n + i] *= corrfac;
	sum += window[ j*n + i];
	if( i != 0 && j!=0 )
	  sum += 3.0 * window[ j*n + i];
	if( (i==0 && j!=0) || (i!=0 && j==0))
	  sum += window[ j*n + i];
      }//for
    }//for
  }//if

  //  printf("sum = %e\n",sum);
  windowfilter( im, window, n, false, i1, j1, i2, j2);

  delete [] window;


  return SUCCESS;

}// mhfilter

/*****************************************************************************/
//
// FUNCTION    : centroid()
// DESCRIPTION : Calculates the centroid and 2nd order moments of the 
//               pixel values around a given input value.
//               
//               
//               
//         
//
// SIDE EFFECTS: UNDER CONSTRUCTION
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//  x0, y0               I                      input coordinates
//  rad                  I                      radius for calculation
//  xc, yc               O                      weighted postion average
//  xyc, x2c, y2c        O                      2nd order moments
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::centroid(float x0, float y0, float rad, float *max,
		       float *xc, float *yc, float *xyc, 
		       float *x2c, float *y2c ) const
{


  if( nx <= 2 || ny <= 2 || x0 < 0 || x0 >= nx || y0 < 0 || y0 >= ny ||
      rad <= 1.2 ){
    //    warning("centroid(): Invalid image or input parameters!");
    return FAILURE;
  }//if


  int i1 = MAX( (unsigned int)(x0 - rad - 0.5) - 1, 0);
  int j1 = MAX( (unsigned int)(y0 - rad - 0.5) - 1, 0);
  int i2 = MIN( (unsigned int)(x0 + rad + 0.5) - 1, nx-1);
  int j2 = MIN( (unsigned int)(y0 + rad + 0.5) - 1, ny-1);
  x0 -= 1.0; // fortran/fist vs C convention
  y0 -= 1.0;


  float pixval, rad2=rad*rad, sum=0.0, sumx=0.0, sumy=0.0;
  int i;
  for(int j=j1 ; j <= j2; j++){
    for( i=i1; i<= i2; i++)
      if( SQR((float)i-x0) + SQR((float)j-y0) < rad2){
        if ( (i + j*nx) < nx*ny ){
	  pixval=pixel[i + j*nx];
	  sum  += pixval;
	  sumx += (i-x0)*pixval;
	  sumy += (j-y0)*pixval;
	}//if
      }//if
  }//for


  if( sum > 0.0 ){
    *xc = sumx/sum + x0;
    *yc = sumy/sum + y0;
  }
  else{
    //    puts("Alarm");
    //    printf("sum=%e %i %i %i %i rad=%.3f\n", sum, i1, j1, i2, j2,rad);
    *xc = x0;
    *yc = y0;
    return FAILURE;
  }//if




  i1 = MAX( (unsigned int)(*xc - rad - 0.5) - 1, 0);
  j1 = MAX( (unsigned int)(*yc- rad - 0.5) - 1, 0);
  i2 = MIN( (unsigned int)(*xc + rad + 0.5) - 1, nx-1);
  j2 = MIN( (unsigned int)(*yc + rad + 0.5) - 1, ny-1);



  *max = 0.0;
  sum = 0.0;
  float sumxx=0.0, sumyy=0.0, sumxy=0.0;
  for(int j=j1 ; j <= j2; j++){
    for( i=i1; i<= i2; i++)
      if( SQR((float)i-x0) + SQR((float)j-y0) < rad2){
        if ( (i + j*nx) < nx*ny ){
	  pixval=pixel[i + j*nx];
	  sum  += pixval;
	  sumxx += SQR(i- *xc)*pixval;
	  sumyy += SQR(j- *yc)*pixval;
	  sumxy += (i- *xc)*(j- *yc)*pixval;
	  *max = MAX( *max, pixval);
        }//if
      }//if
  }//for


  if( sum > 0.0 ){
    *xyc = sumxy/sum;  
    *x2c = sumxx/sum;  
    *y2c = sumyy/sum;
  }
  else{
    //    puts("Alarm");
    *xyc = *x2c = *y2c = 0.0;
    return FAILURE;
  }//if
  *xc += 1.0; // fortan/fits vs C convention
  *yc += 1.0;


  return SUCCESS;
}//centroid( )




/*****************************************************************************/
//
// FUNCTION    : addgaussian()
// DESCRIPTION : adds a gaussian with a total integration of value "norm" 
//               and specified width to the image
// 
//
// REMARKS     :
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//    x,y               I                      pixel position
//    norm              I                      total integration
//   sigma_x,sigma_y    I                      x/y width
//   rot                I                      rotation (in degrees)
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::addgaussian(float x, float y, float norm,
			  float sigmax, float sigmay, float rot)
{
  if( sigmax <= 0.0 || sigmay <= 0.0 || x <= 0.0 || y <= 0 || 
      x >= nx || y >= ny){
    //    warning("addgaussian(): Error in sigmax, sigmay input values!");
    return FAILURE;
  }//if

  float radius = 4.0 * MAX(sigmax, sigmay);
  int i1 = (int)(x - radius - 0.5)-1,  j1=(int)(y - radius -0.5)-1;
  int i2 = (int)(x + radius + 0.5)-1, j2=(int)(y + radius + 0.5)-1;

  i1 = MAX(i1, 0);
  j1 = MAX(j1, 0);
  i2 = MIN(i2, (int)(nx-1));
  j2 = MIN(j2, (int)(nx-1));

  x -= 1.0; // FITS 2 C++ convention for pixels
  y -= 1.0; // FITS 2 C++ convention for pixels
  norm /= 2.0 * M_PI * sigmax * sigmay;
  rot *= M_PI/180.0; //convert to radians


  float x2, y2, x3, y3, sinrot=sin(rot), cosrot=cos(rot);

  int i;
  for( int j = j1 ; j <= j2 ; j++ )
    for( i = i1 ; i <= i2 ; i++ ){

      x2 = (i-x)/sigmax;
      y2 = (j-y)/sigmay;
      if( rot != 0.0 && sigmax != sigmay ){
	x3 = x2*cosrot - y2*sinrot;
	y3 = x2*sinrot + y2*cosrot;
	x2 = x3;
	y2 = y3;
      }
      pixel[i + j*nx] += norm * exp( -0.5*(x2*x2 + y2*y2) );

    }//for

  return SUCCESS;
}// addgaussian()


/*****************************************************************************/
//
// FUNCTION    : exposuremask()
// DESCRIPTION : Uses the exposure map to create an image containing pixels=1,
//               where photons can be detected and 0.0 elsewhere.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :
//
/*****************************************************************************/
Status Image::exposuremask(const Image &expmap)
{
  if( expmap.nx < 1 || expmap.ny < 1 ){
    //    warning("exposuremask(): Invalid exposure map!");
    return FAILURE;
  }//if

  *this = expmap;

  for( unsigned long i = 0; i < (unsigned long)nx*ny ; i++)
    pixel[i] = (expmap.pixel[i] > 0.5 ? 1.0 : 0.0);

  return SUCCESS;

}//exposuremask()
